﻿Imports System
Namespace ChiSquare
  Friend Class ChiSquareProgram
	Shared Sub Main(ByVal args() As String)
	  Try
		Console.WriteLine(VBCrLf & "Begin Chi-square test using C# demo" & VBCrLf)
		Console.WriteLine("Goal is to see if one die from a set of dice is biased or not" & VBCrLf)

		Dim observed() As Integer = { 20, 28, 12, 32, 22, 36 }

		Console.WriteLine(VBCrLf & "Starting chi-square test")
		Dim p As Double = ChiSquareTest(observed)
		Console.WriteLine(VBCrLf & "Chi-square test complete")

		Dim crit As Double = 0.05
		If p < crit Then
		  Console.WriteLine(VBCrLf & "Because p-value is below critical value of " & crit.ToString("F2"))
		  Console.WriteLine("the null hypothsis is rejected and we conclude")
		  Console.WriteLine("the data is unlikely to have happened by chance.")
		Else
		  Console.WriteLine(VBCrLf & "Because p-value is not below critical value of " & crit.ToString("F2"))
		  Console.WriteLine("the null hypothsis is accepted (not rejected) and we conclude")
		  Console.WriteLine("the observed data could have happened by chance.")
		End If

		Console.WriteLine(VBCrLf & "End" & VBCrLf)
		Console.ReadLine()
	  Catch ex As Exception
		Console.WriteLine(ex.Message)
		Console.ReadLine()
	  End Try

	End Sub ' Main

	Private Shared Sub ShowVector(ByVal v() As Integer)
	  For i = 0 To v.Length - 1
		Console.Write(v(i) & " ")
	  Next i
	  Console.WriteLine("")
	End Sub

	Private Shared Function ChiSquareTest(ByVal observed() As Integer) As Double
	  Console.WriteLine("Observed frequencies are: ")
	  ShowVector(observed)

	  Dim x As Double = ChiSquareStatistic(observed)
	  Console.WriteLine(VBCrLf & "X-squared = " & x.ToString("F2"))

	  Dim df As Integer = observed.Length - 1
	  Console.WriteLine(VBCrLf & "df = " & df)

	  Dim p As Double = ChiSquareProb(x, df)
	  Console.WriteLine(VBCrLf & "p-value = " & p.ToString("F6"))

	  Return p
	End Function

	Private Shared Function ChiSquareStatistic(ByVal observed() As Integer) As Double
	  Dim sumObs As Double = 0.0
	  For i = 0 To observed.Length - 1
		sumObs += observed(i)
	  Next i
	  Dim expected As Double = CInt(Fix(sumObs / observed.Length))

	  Dim result As Double = 0.0
	  For i = 0 To observed.Length - 1
		result = result +((observed(i) - expected) * (observed(i) - expected)) / expected
	  Next i
	  Return result
	End Function

	Public Shared Function ChiSquareProb(ByVal x As Double, ByVal df As Integer) As Double
	  ' x = a computed chi-square value. df = degrees of freedom.
	  ' output = prob. the x value occurred by chance.
	  ' so, for example, if result < 0.05 there is only a 5% chance
	  ' that the x value ocurred by chance and therefore 
	  ' we conclude that the actual data which produced x is
	  ' NOT the same as the expected data.
	  ' this function can be used to create a ChiSquareTest procedure.
	  ' ACM Algorithm 299 and update ACM TOMS June 1985.
	  ' Uses custom Exp() function below.

	  If x <= 0.0 OrElse df < 1 Then
		Throw New Exception("parameter x must be positive " & "and parameter df must be 1 or greater in ChiSquaredProb()")
	  End If

	  Dim a As Double = 0.0 ' 299 variable names
	  Dim y As Double = 0.0
	  Dim s As Double = 0.0
	  Dim z As Double = 0.0
	  Dim e As Double = 0.0
	  Dim c As Double

	  Dim even As Boolean ' is df even?

	  a = 0.5 * x
	  If df Mod 2 = 0 Then
		  even = True
		  Else
			  even = False
		  End If

	  If df > 1 Then ' ACM update remark (4)
		  y = Exp(-a)
	  End If

	  If even = True Then
		  s = y
		  Else
			  s = 2.0 * Gauss(-Math.Sqrt(x))
		  End If

	  If df > 2 Then
		x = 0.5 * (df - 1.0)
		If even = True Then
			z = 1.0
			Else
				z = 0.5
			End If
		If a > 40.0 Then ' ACM remark (5)
		  If even = True Then ' log(sqrt(pi))
			  e = 0.0
			  Else
				  e = 0.5723649429247000870717135
			  End If
		  c = Math.Log(a) ' log base e
		  Do While z <= x
			e = Math.Log(z) + e
			s = s + Exp(c * z - a - e) ' ACM update remark (6)
			z = z + 1.0
		  Loop
		  Return s ' a > 40.0
		Else
		  If even = True Then ' (1 / sqrt(pi))
			  e = 1.0
			  Else
				  e = 0.5641895835477562869480795 / Math.Sqrt(a)
			  End If
		  c = 0.0
		  Do While z <= x
			e = e * (a / z) ' ACM update remark (7)
			c = c + e
			z = z + 1.0
		  Loop
		  Return c * y + s
		End If ' df > 2
	  Else
		Return s
	  End If
	End Function ' ChiSquare()

	Private Shared Function Exp(ByVal x As Double) As Double ' ACM update remark (3)
	  If x < -40.0 Then ' 40.0 is a magic number. ACM update remark (8)
		Return 0.0
	  Else
		Return Math.Exp(x)
	  End If
	End Function

	Public Shared Function Gauss(ByVal z As Double) As Double
	  ' input = z-value (-inf to +inf)
	  ' output = p under Normal curve from -inf to z
	  ' e.g., if z = 0.0, function returns 0.5000
	  ' ACM Algorithm #209
	  Dim y As Double ' 209 scratch variable
	  Dim p As Double ' result. called 'z' in 209
	  Dim w As Double ' 209 scratch variable

	  If z = 0.0 Then
		p = 0.0
	  Else
		y = Math.Abs(z) / 2
		If y >= 3.0 Then
		  p = 1.0
		ElseIf y < 1.0 Then
		  w = y * y
		  p = ((((((((0.000124818987 * w - 0.001075204047) * w + 0.005198775019) * w - 0.019198292004) * w + 0.059054035642) * w - 0.151968751364) * w + 0.319152932694) * w - 0.531923007300) * w + 0.797884560593) * y * 2.0
		Else
		  y = y - 2.0
		  p = (((((((((((((-0.000045255659 * y + 0.000152529290) * y - 0.000019538132) * y - 0.000676904986) * y + 0.001390604284) * y - 0.000794620820) * y - 0.002034254874) * y + 0.006549791214) * y - 0.010557625006) * y + 0.011630447319) * y - 0.009279453341) * y + 0.005353579108) * y - 0.002141268741) * y + 0.000535310849) * y + 0.999936657524
		End If
	  End If

	  If z > 0.0 Then
		Return (p + 1.0) / 2
	  Else
		Return (1.0 - p) / 2
	  End If
	End Function ' Gauss()

  End Class ' Program class

End Namespace ' ns
